
local eqg = require "luaeqg"
local lfs = require "lfs"

local list = iup.list{visiblelines = 10, expand = "VERTICAL", visiblecolumns = 16,
	sort = "YES", multiple = "YES"}

local filter = iup.text{visiblecolumns = 15, value = ""}
file_filter = filter

function filter:valuechanged_cb()
	local dir = open_dir
	if dir then
		FilterFileList(dir)
	end
end

local ipairs = ipairs
local pcall = pcall
local by_name = {}

function UpdateFileList(path)
	eqg.CloseDirectory(open_dir)
	open_path = path
	local s, dir = pcall(eqg.LoadDirectory, path)
	if not s then
		error_popup(dir)
		return
	end
	open_dir = dir
	FilterFileList(dir)
end

function FilterFileList(dir)
	list[1] = nil
	list.autoredraw = "NO"
	local f = filter.value
	if f:len() > 0 then
		f = f:gsub("%.", "%%%.")
		if f:find("%%", -1) then
			f = f .. "%"
		end
		local n = 1
		for i, ent in ipairs(dir) do
			ent.pos = i
			local name = ent.name
			by_name[name] = ent
			if name:find(f) then
				list[n] = name
				n = n + 1
			end
		end
	else
		for i, ent in ipairs(dir) do
			ent.pos = i
			local name = ent.name
			list[i] = name
			by_name[name] = ent
		end
	end
	list.autoredraw = "YES"
end

function SetTop(name)
	local n = list.count
	for i = 1, n do
		if list[i] == name then
			list.topitem = i
			list:action(name, i, 1)
			return
		end
	end
end

local selection
function list:action(str, pos, state)
	if state == 1 then
		local sel = by_name[str]
		if selection ~= sel then
			selection = sel
			local s, err = pcall(eqg.OpenEntry, sel)
			if s then
				SetCurImage(sel.ptr, sel.inflated_len)
				iup.Update(canvas)
			else
				error_popup(err)
			end
		end
	end
end

local find = string.find
function Export()
	local dlg = iup.filedlg{title = "Export to...", dialogtype = "DIR"}
	iup.Popup(dlg)
	if dlg.status == "0" then
		local path = dlg.value
		local val = list.value
		if path and val then
			local i = 0
			while true do
				i = find(val, "%+", i + 1)
				if not i then break end
				local str = list[i]
				local outpath = path .."\\".. str
				local data = by_name[str]
				local s, err = pcall(eqg.OpenEntry, data)
				if s then
					s, err = pcall(eqg.ExportFile, outpath, data)
					if not s then
						error_popup(err)
						break
					end
				else
					error_popup(err)
					break
				end
			end
		end
	end
	iup.Destroy(dlg)
end

function Import()
	local dlg = iup.filedlg{title = "Select files to import...",
		dialogtype = "FILE", multiplefiles = "YES"}
	iup.Popup(dlg)
	if dlg.status == "0" then
		local path = dlg.value
		local dir = open_dir
		if path and dir then
			if path:match("%|") then
				--there are multiple files to import
				local base = path:match("[^%|]+") .. "\\"
				for name in path:gmatch("%|([^%|]+)") do
					local n = name:lower()
					local pos = by_name[name]
					pos = pos and pos.pos or #dir + 1
					local s, err = pcall(eqg.ImportFile, base .. n, n, dir, pos)
					if not s then
						error_popup(err)
						break
					end
				end
				local s, err = pcall(eqg.WriteDirectory, open_path, dir)
				if s then
					UpdateFileList(open_path)
				else
					error_popup(err)
				end
			else
				--only one file
				local name = string.lower(path:match("[^\\]+$"))
				if name then
					local pos = by_name[name]
					pos = pos and pos.pos or #dir + 1
					local s, err = pcall(eqg.ImportFile, path, name, dir, pos)
					if s then
						s, err = pcall(eqg.WriteDirectory, open_path, dir)
						if s then
							UpdateFileList(open_path)
							iup.Destroy(dlg)
							return
						end
					end
					error_popup(err)
				end
			end
		end
	end
	iup.Destroy(dlg)
end

function Delete()
	local dlg = iup.messagedlg{buttondefault = "2", title = "Confirm Deletion",
		buttons = "OKCANCEL", dialogtype = "WARNING", value = "Are you sure you want to delete these files?"}
	iup.Popup(dlg)
	if dlg.buttonresponse == "2" then return end
	local dir = open_dir
	local path = open_path
	local val = list.value
	if dir and path and val then
		local i = 0
		local remove = {}
		while true do
			i = find(val, "%+", i + 1)
			if not i then break end
			table.insert(remove, by_name[list[i]].pos)
		end
		table.sort(remove)
		local adj = 0
		for _, pos in ipairs(remove) do
			table.remove(dir, pos - adj)
			adj = adj + 1
		end
		local s, err = pcall(eqg.WriteDirectory, path, dir)
		if s then
			UpdateFileList(path)
			iup.Destroy(dlg)
			return
		end
		error_popup(err)
	end
	iup.Destroy(dlg)
end

function list:button_cb(button, pressed, x, y)
	if button == iup.BUTTON3 then
		local pos = iup.ConvertXYToPos(self, x, y)
		if pos ~= -1 then
			if pressed == 0 then
				local has = list.value:match("%+") and "YES" or "NO"
				local mx, my = iup.GetGlobal("CURSORPOS"):match("(%d+)x(%d+)")
				local menu = iup.menu{
					iup.item{title = "Export Selected Files...", action = Export, active = has},
					iup.item{title = "Import Files...", action = Import},
					iup.separator{},
					iup.item{title = "Delete Selected Files", action = Delete, active = has},
				}
				iup.Popup(menu, mx, my)
				iup.Destroy(menu)
			end
		end
	end
end

return iup.vbox{iup.hbox{iup.label{title = "Filter"}, filter; alignment = "ACENTER", gap = 5}, list;
	alignment = "ACENTER", gap = 5}
