
#include <lua.hpp>
#include <zlib.h>
#include <map>
#include <cstdio>
#include <cstring>
#include <cstdint>

typedef std::uint8_t byte;
typedef std::uint32_t uint32;

class Buffer
{
public:
	Buffer(bool huge = false);
	void Add(byte* data, uint32 len);
	byte* Take();
	uint32 GetLen();
private:
	uint32 mLen;
	uint32 mCap;
	byte* mData;
};

struct Header
{
	uint32 offset;
	char magic[4];
	uint32 unknown;
	static const uint32 SIZE = 12;
};

struct BlockHeader
{
	uint32 deflatedLen;
	uint32 inflatedLen;
};

struct DirEntry
{
	uint32 crc;
	uint32 offset;
	uint32 inflatedLen;
	static const uint32 SIZE = 12;
};

struct DirMeta
{
	DirEntry* dir;
	uint32 deflatedLen;
};

uint32 crc_table[256];
void GenerateCRCTable()
{
	for (uint32 i = 0; i < 256; ++i)
	{
		uint32 v = i << 24;
		for (uint32 j = 0; j < 8; ++j)
		{
			if ((v & 0x80000000) != 0)
			{
				static const uint32 polynomial = 0x04C11DB7;
				v = (v << 1) ^ polynomial;
			}
			else
			{
				v <<= 1;
			}
		}
		crc_table[i] = v;
	}
}

uint32 CalcCRC(byte* ptr, uint32 len)
{
	uint32 v = 0;
	while (len > 0)
	{
		uint32 i = ((v >> 24) ^ *ptr) & 0xFF;
		v = (v << 8) ^ crc_table[i];
		++ptr;
		--len;
	}
	return v;
}

Buffer::Buffer(bool huge) : mLen(0)
{
	if (huge)
	{
		mData = new byte[1 << 22]; //4MB
		mCap = 1 << 22;
	}
	else
	{
		mData = new byte[8192];
		mCap = 8192;
	}
}

void Buffer::Add(byte* data, uint32 len)
{
	uint32 newlen = mLen + len;
	if (newlen >= mCap)
	{
		while (newlen >= mCap)
			mCap <<= 1;
		byte* add = new byte[mCap];
		memcpy(add, mData, mLen);
		delete[] mData;
		mData = add;
	}
	memcpy(&mData[mLen], data, len);
	mLen = newlen;
}

byte* Buffer::Take()
{
	byte* ret = mData;
	mData = nullptr;
	return ret;
}

uint32 Buffer::GetLen()
{
	return mLen;
}

byte* DecompressDirEntry(lua_State* L, byte* data, uint32 inflatedLen);

int LoadDirectory(lua_State* L)
{
	const char* path = luaL_checkstring(L, 1);

	FILE* fp = fopen(path, "rb");
	if (fp == nullptr)
		return luaL_argerror(L, 1, "could not open specified path");

	lua_newtable(L); //to return

	fseek(fp, 0, SEEK_END);
	uint32 len = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	byte* data = (byte*)lua_newuserdata(L, len);
	fread(data, sizeof(byte), len, fp);
	fclose(fp);

	lua_setfield(L, -2, "raw_data");

	Header* h = (Header*)data;
	if (h->magic[0] != 'P' || h->magic[1] != 'F' || h->magic[2] != 'S' || h->magic[3] != ' ')
	{
		lua_pop(L, 1);
		return luaL_argerror(L, 1, "file is not a valid s3d/eqg directory file");
	}

	uint32 num_entries = *(uint32*)&data[h->offset];
	uint32 pos = h->offset + sizeof(uint32);
	uint32 high_offset = 0;
	DirEntry* last;

	//Directory Entries are listed in order of CRC, but the list of file names are
	//in order of entry offset - load entries into a map sorted by offset

	std::map<uint32, DirMeta> entryMap;
	for (uint32 i = 0; i < num_entries; ++i)
	{
		DirEntry* entry = (DirEntry*)&data[pos];
		pos += DirEntry::SIZE;
		if (entry->offset > high_offset)
		{
			high_offset = entry->offset;
			last = entry;
		}

		DirMeta dm;
		dm.dir = entry;
		//record deflatedLen
		uint32 dpos = entry->offset;
		uint32 dlen = 0;
		uint32 ilen = 0;
		while (ilen < entry->inflatedLen)
		{
			BlockHeader* bh = (BlockHeader*)&data[dpos];
			uint32 shift = sizeof(BlockHeader) + bh->deflatedLen;
			dpos += shift;
			dlen += shift;
			ilen += bh->inflatedLen;
		}
		dm.deflatedLen = dlen;

		entryMap[entry->offset] = dm;
	}

	byte* name_data = DecompressDirEntry(L, &data[last->offset], last->inflatedLen);
	uint32 n = *(uint32*)name_data;
	pos = sizeof(uint32);
	auto itr = entryMap.begin();

	for (uint32 i = 1; i <= n; ++i)
	{
		//pascal-style strings with 32bit len field (and a null terminator just cuz)
		uint32 len = *(uint32*)&name_data[pos];
		pos += sizeof(uint32);
		char* name = (char*)&name_data[pos];
		pos += len;

		DirMeta& entry = itr->second;
		DirEntry* dir = entry.dir;

		//set lua data
		lua_pushinteger(L, i);
		lua_newtable(L);

		lua_pushstring(L, name);
		lua_setfield(L, -2, "name");
		lua_pushinteger(L, dir->crc);
		lua_setfield(L, -2, "crc");
		lua_pushinteger(L, dir->inflatedLen);
		lua_setfield(L, -2, "inflated_len");
		lua_pushlightuserdata(L, &data[dir->offset]);
		lua_setfield(L, -2, "ptr");
		lua_pushinteger(L, entry.deflatedLen);
		lua_setfield(L, -2, "deflated_len");
		//decompressed = nil

		lua_settable(L, -3);
		++itr;
	}

	delete[] name_data;

	return 1;
}

void Decompress(lua_State* L, byte* dst, uint32 dlen, byte* src, uint32 slen)
{
	z_stream z;

	z.zalloc = nullptr;
	z.zfree = nullptr;
	z.opaque = nullptr;

	z.next_in = src;
	z.avail_in = slen;
	z.next_out = dst;
	z.avail_out = dlen;

	if (inflateInit(&z) != Z_OK || inflate(&z, Z_NO_FLUSH) != Z_STREAM_END || inflateEnd(&z) != Z_OK)
		luaL_error(L, "ZLib inflation failed");
}

uint32 Compress(lua_State* L, byte* dst, uint32 dlen, byte* src, uint32 slen)
{
	z_stream z;

	z.zalloc = nullptr;
	z.zfree = nullptr;
	z.opaque = nullptr;

	z.next_in = src;
	z.avail_in = slen;
	z.next_out = dst;
	z.avail_out = dlen;

	if (deflateInit(&z, 9) != Z_OK || deflate(&z, Z_FINISH) != Z_STREAM_END || deflateEnd(&z) != Z_OK)
		luaL_error(L, "ZLib deflation failed");

	return z.total_out;
}

byte* DecompressDirEntry(lua_State* L, byte* data, uint32 inflatedLen)
{
	uint32 read = 0;
	uint32 pos = 0;
	Buffer buf;

	while (read < inflatedLen)
	{
		BlockHeader* bh = (BlockHeader*)&data[pos];
		pos += sizeof(BlockHeader);

		byte decompressed[8192];
		Decompress(L, decompressed, 8192, &data[pos], bh->deflatedLen);
		read += bh->inflatedLen;
		pos += bh->deflatedLen;
		buf.Add(decompressed, bh->inflatedLen);
	}

	return buf.Take();
}

void CompressDirEntry(lua_State* L, Buffer& buf, byte* data, uint32 inflatedLen, uint32& file_pos)
{
	Buffer b;
	uint32 read = 0;
	while (inflatedLen > 0)
	{
		static const uint32 MAX_BLOCK = 8192;
		uint32 r = (inflatedLen < MAX_BLOCK) ? inflatedLen : MAX_BLOCK;

		BlockHeader bh;
		bh.inflatedLen = r;

		byte compressed[MAX_BLOCK << 1];
		uint32 len = Compress(L, compressed, MAX_BLOCK << 1, &data[read], r);
		bh.deflatedLen = len;

		b.Add((byte*)&bh, sizeof(BlockHeader));
		b.Add(compressed, len);

		inflatedLen -= r;
		read += r;
	}

	byte* add = b.Take();
	uint32 len = b.GetLen();
	buf.Add(add, len);
	delete[] add;

	file_pos += len;
}

uint32 GetInt(lua_State* L, int pos, const char* name)
{
	uint32 val = 0;
	lua_getfield(L, pos, name);
	val |= lua_tointeger(L, -1);
	lua_pop(L, 1);
	return val;
}

int WriteDirectory(lua_State* L)
{
	const char* path = luaL_checkstring(L, 1);
	if (!lua_istable(L, 2))
		return luaL_argerror(L, 2, "expected directory data table");

	FILE* fp = fopen(path, "wb+");
	if (fp == nullptr)
		return luaL_argerror(L, 1, "could not open file for writing");

	//header
	//entries, in any order
	//names entry, in order of entry offsets
	//entry offset and crc list, in order of crc

	Header header;
	header.magic[0] = 'P';
	header.magic[1] = 'F';
	header.magic[2] = 'S';
	header.magic[3] = ' ';
	header.unknown = 131072;

	uint32 file_pos = Header::SIZE;
	//multiple entries having the same crc is pretty unlikely... but just in case
	std::multimap<uint32, DirEntry> entryMap;
	Buffer buf(true);
	Buffer name_buf;

	int n = lua_objlen(L, 2);
	name_buf.Add((byte*)&n, sizeof(int));
	for (int i = 1; i <= n; ++i)
	{
		lua_pushinteger(L, i);
		lua_gettable(L, -2);

		//entry table on stack
		//record name in name list (can't compress until we have them all)
		lua_getfield(L, -1, "name");
		uint32 len = lua_objlen(L, -1) + 1;
		name_buf.Add((byte*)&len, sizeof(uint32));
		name_buf.Add((byte*)lua_tostring(L, -1), len);
		lua_pop(L, 1);

		DirEntry ent;
		ent.crc = GetInt(L, -1, "crc");
		ent.offset = file_pos;
		ent.inflatedLen = GetInt(L, -1, "inflated_len");
		entryMap.insert(std::make_pair(ent.crc, ent));

		lua_getfield(L, -1, "decompressed");
		if (lua_toboolean(L, -1))
		{
			//decompressed entry, compress
			lua_getfield(L, -2, "ptr");
			byte* ptr = (byte*)lua_touserdata(L, -1);
			lua_pop(L, 1);

			CompressDirEntry(L, buf, ptr, ent.inflatedLen, file_pos);
		}
		else
		{
			//still compressed, can just copy it in
			len = GetInt(L, -2, "deflated_len");

			lua_getfield(L, -2, "ptr");
			byte* ptr = (byte*)lua_touserdata(L, -1);
			lua_pop(L, 1);

			buf.Add(ptr, len);

			file_pos += len;
		}

		lua_pop(L, 2); //bool and entry table
	}

	//compress names entry so we can tell the header the offset after it
	DirEntry ent;
	ent.crc = 0x61580AC9; //always this
	ent.offset = file_pos;
	ent.inflatedLen = name_buf.GetLen();
	entryMap.insert(std::make_pair(ent.crc, ent));

	Buffer add_names;
	byte* b = name_buf.Take();
	CompressDirEntry(L, add_names, b, name_buf.GetLen(), file_pos);
	delete[] b;

	header.offset = file_pos;

	//finally, start writing to file
	fwrite(&header, Header::SIZE, 1, fp);

	//write compressed entries
	b = buf.Take();
	fwrite(b, sizeof(byte), buf.GetLen(), fp);
	delete[] b;

	//write file names entry
	b = add_names.Take();
	fwrite(b, sizeof(byte), add_names.GetLen(), fp);
	delete[] b;

	//write offset and crc list in order of crc
	uint32 len = entryMap.size();
	fwrite(&len, sizeof(uint32), 1, fp);
	for (auto itr = entryMap.begin(); itr != entryMap.end(); ++itr)
	{
		DirEntry& entry = itr->second;
		fwrite(&entry, DirEntry::SIZE, 1, fp);
	}

	fclose(fp);

	return 0;
}

int OpenDirectoryEntry(lua_State* L)
{
	if (!lua_istable(L, 1))
		return luaL_argerror(L, 1, "expected data table");

	lua_getfield(L, 1, "decompressed");
	if (!lua_toboolean(L, -1))
	{
		lua_pushboolean(L, 1);
		lua_setfield(L, 1, "decompressed");

		uint32 len = GetInt(L, 1, "inflated_len");
		lua_getfield(L, 1, "ptr");
		byte* ptr = (byte*)lua_touserdata(L, -1);
		lua_pop(L, 1);

		byte* add = DecompressDirEntry(L, ptr, len);
		lua_pushlightuserdata(L, add);
		lua_setfield(L, 1, "ptr");
	}

	return 0;
}

int CloseDirectory(lua_State* L)
{
	if (lua_istable(L, 1))
	{
		int n = lua_objlen(L, 1);
		for (int i = 1; i <= n; ++i)
		{
			lua_pushinteger(L, i);
			lua_gettable(L, -2);

			lua_getfield(L, -1, "decompressed");
			if (lua_toboolean(L, -1))
			{
				lua_getfield(L, -2, "ptr");
				byte* ptr = (byte*)lua_touserdata(L, -1);
				lua_pop(L, 1);
				delete[] ptr;
			}

			lua_pop(L, 2); //tbl, bool
		}
	}
	return 0;
}

int ImportDirectoryFile(lua_State* L)
{
	const char* path = luaL_checkstring(L, 1);
	const char* name = luaL_checkstring(L, 2); //lowercased on the lua side
	if (!lua_istable(L,3))
		return luaL_argerror(L, 3, "expected directory data table");
	int pos = luaL_checkinteger(L, 4);

	FILE* fp = fopen(path, "rb");
	if (fp == nullptr)
		return luaL_argerror(L, 1, "could not open file to import");

	fseek(fp, 0, SEEK_END);
	uint32 len = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	byte* data = new byte[len];
	fread(data, sizeof(byte), len, fp);
	fclose(fp);

	lua_pushinteger(L, pos);
	lua_newtable(L);

	int i = lua_objlen(L, 2) + 1;

	lua_pushstring(L, name);
	lua_setfield(L, -2, "name");
	lua_pushinteger(L, CalcCRC((byte*)name, i));
	lua_setfield(L, -2, "crc");
	lua_pushboolean(L, 1);
	lua_setfield(L, -2, "decompressed");
	lua_pushinteger(L, len);
	lua_setfield(L, -2, "inflated_len");
	lua_pushlightuserdata(L, data);
	lua_setfield(L, -2, "ptr");

	lua_settable(L, 3);

	return 0;
}

int ExportDirectoryFile(lua_State* L)
{
	const char* out_path = luaL_checkstring(L, 1);
	if (!lua_istable(L, 2))
		return luaL_argerror(L, 2, "expected data table");

	//data has already been decompressed
	lua_getfield(L, 2, "inflated_len");
	int len = lua_tointeger(L, -1);
	lua_getfield(L, 2, "ptr");
	byte* ptr = (byte*)lua_touserdata(L, -1);
	lua_pop(L, 2);

	FILE* fp = fopen(out_path, "wb+");
	if (fp == nullptr)
		return luaL_argerror(L, 1, "could not open file path for writing");

	fwrite(ptr, sizeof(byte), len, fp);

	fclose(fp);

	return 0;
}

int LuaCalcCRC(lua_State* L)
{
	const char* str = luaL_checkstring(L, 1);
	uint32 len = lua_objlen(L, 1) + 1;

	lua_pushinteger(L, CalcCRC((byte*)str, len));
	return 1;
}

static const luaL_Reg funcs[] = {
	{"LoadDirectory", LoadDirectory},
	{"CloseDirectory", CloseDirectory},
	{"WriteDirectory", WriteDirectory},
	{"OpenEntry", OpenDirectoryEntry},
	{"ImportFile", ImportDirectoryFile},
	{"ExportFile", ExportDirectoryFile},
	{"CalcCRC", LuaCalcCRC},
	{nullptr, nullptr}
};

extern "C"
int __declspec(dllexport) luaopen_luaeqg(lua_State* L)
{
	GenerateCRCTable();
	lua_newtable(L);
	luaL_register(L, nullptr, funcs);
	lua_pushvalue(L, -1);
	lua_setglobal(L, "luaeqg");
	return 1;
}
